const fs = require('fs')
const path = require('path');
const chalk = require('chalk');
const webpack = require('webpack');
const filesize = require('filesize');
const { appComponents, appBuild } = require('./paths')
const { checkComponentFiles } = require('./utils')
const getWebpackConfig = require('./webpack.config.build');

function canReadAsset(asset) {
    return (
        /\.(js|css)$/.test(asset) &&
        !/service-worker\.js/.test(asset) &&
        !/precache-manifest\.[0-9a-f]+\.js/.test(asset)
    );
}

module.exports = (name, option = {}) => {
    const { unique = true } = option;
    return new Promise((resolve, reject) => {
        // 检查组件是否存在
        // 开始打包
        // 打包成功 返回bundle路径

        const sourcePath = path.join(appComponents, name);

        let easyvcid;
        if (unique) {
            // 生成组件唯一标识
            const easyvc = path.join(sourcePath, '.easyvc');
            if (fs.existsSync(easyvc)) {
                easyvcid = fs.readFileSync(easyvc, { encoding: 'utf8' });
            }
            if (!easyvcid) {
                easyvcid = Array.from({ length: 32 }).map(() => Math.random().toString(32).slice(2, 3)).join('');
                fs.writeFileSync(easyvc, easyvcid, { encoding: 'utf8' });
            }
        }

        // 重写base的modulename
        const { base } = checkComponentFiles(sourcePath);
        base.filename = `${base.module_name}`;
        base.module_name = `${base.module_name}${unique ? '.' + easyvcid : ''}`;
        const compiler = webpack(getWebpackConfig(base, { uniqueId: easyvcid }))

        console.log('Creating an optimized production build...');

        compiler.run((err, stats) => {
            const message = stats.toJson({ all: true, warnings: true, errors: true });

            const errors = err || message.errors;
            if (errors.length) {
                console.log(chalk.red('Failed to compile.\n'));
                console.log(chalk.red((errors) + '\n'));
                process.exit(1);
            }

            if (message.warnings.length) {
                console.log(chalk.yellow('Compiled with warnings.\n'));
                console.log(message.warnings.join('\n\n'));
            } else {
                // successful
                console.log(chalk.green('Compiled successfully.'));
                console.log(`Time: ${message.time / 1000}s`);
                console.log(`Built at: ${new Date(message.builtAt)}\n`);
                console.log('File sizes:\n');

                message.assets
                    .filter(asset => canReadAsset(asset.name))
                    .map(asset => {
                        var fileContents = fs.readFileSync(path.resolve(appBuild, asset.name));
                        return {
                            folder: path.join(
                                path.basename(appBuild),
                                path.dirname(asset.name)
                            ),
                            name: path.basename(asset.name),
                            size: filesize(fileContents.length),
                        };
                    }).sort((a, b) => b.size - a.size).map(asset => {
                        console.log(
                            '  ' +
                            asset.size +
                            '  ' +
                            chalk.dim(asset.folder + path.sep) +
                            chalk.cyan(asset.name)
                        );
                    });

                console.log();
            }
            resolve({ file: path.join(message.outputPath, message.assetsByChunkName.main), base, sourcePath: sourcePath })
        });


    })
}