/**
 * 给导出的组件添加static方法获取配置文件
 */
const path = require('path')
const glob = require('glob')
const fs = require('fs');

const parser = require('@babel/parser').parse;
const traverse = require("@babel/traverse").default;
const generate = require("@babel/generator").default;
const t = require("@babel/types");


function getOldDimension(width, height, hasUndeline) {
    if (hasUndeline) {
        return {
            "_name": "dimension",  // 定义位置宽高
            "_displayName": "位置尺寸",
            "_value": [
                {
                    "_name": "chartPosition",
                    "_displayName": "图表位置",
                    "_value": [
                        {
                            "_name": "left",
                            "_displayName": "X轴坐标",
                            "_value": 100  // 开发环境默认显示位置
                        },
                        {
                            "_name": "top",
                            "_displayName": "Y轴坐标",
                            "_value": 100  // 开发环境默认显示位置
                        }
                    ]
                },
                {
                    "_name": "chartDimension",
                    "_displayName": "图表尺寸",
                    "_value": [
                        {
                            "_name": "width",
                            "_displayName": "宽度",
                            "_value": width  // 组件默认宽
                        },
                        {
                            "_name": "height",
                            "_displayName": "高度",
                            "_value": height  // 组件默认高
                        }
                    ]
                }
            ]
        }
    }
    return {
        "name": "dimension",  // 定义位置宽高
        "displayName": "位置尺寸",
        "value": [
            {
                "name": "chartPosition",
                "displayName": "图表位置",
                "value": [
                    {
                        "name": "left",
                        "displayName": "X轴坐标",
                        "value": 100  // 开发环境默认显示位置
                    },
                    {
                        "name": "top",
                        "displayName": "Y轴坐标",
                        "value": 100  // 开发环境默认显示位置
                    }
                ]
            },
            {
                "name": "chartDimension",
                "displayName": "图表尺寸",
                "value": [
                    {
                        "name": "width",
                        "displayName": "宽度",
                        "value": width  // 组件默认宽
                    },
                    {
                        "name": "height",
                        "displayName": "高度",
                        "value": height  // 组件默认高
                    }
                ]
            }
        ]
    }
}

module.exports = function (content, map, meta) {

    var callback = this.async();
    const query = this.query || {};

    const modulePath = this.context;
    let file = content;

    // 兼容旧的写法
    if (~file.indexOf('getDefaultConfig')) {
        return callback(null, file);
    }
    const configPath = `${modulePath}/config`;
    // 添加watcher依赖
    this.addContextDependency(configPath)
    glob(`*.data.json`, {
        cwd: configPath
    }, function (err, matches) {
        if (err) {
            callback(`main.json or main.data.json cannot be found in ${configPath}`)
            return;
        }

        try {
            // 读取配置文件
            const main = {};
            const subConfigs = [];
            const subDatas = [];
            const names = new Set();
            if (matches.map(filename => filename.split('.')[0]).some((name, i) => {
                const path1 = path.join(configPath, name + '.json'),
                    path2 = path.join(configPath, name + '.data.json')
                let config = fs.readFileSync(path1, { encoding: 'utf8' })
                try { config = JSON.parse(config); } catch (error) {
                    error.message += ` in ${path1}`;
                    callback(error)
                    return true
                }
                let data = fs.readFileSync(path2, { encoding: 'utf8' });
                try { data = JSON.parse(data); } catch (error) {
                    error.message += ` in ${path2}`;
                    callback(error)
                    return true;
                }
                if (names.has(config.base.module_name)) {
                    callback(new Error(`base.module_name ${config.base.module_name} was duplicated in ${configPath}`));
                    return true;
                } else {
                    names.add(config.base.module_name)
                }
                if (name === 'main') {
                    main.config = config;
                    main.data = data;
                    return;
                }
                const id = Math.random().toString(16).slice(2)
                config.id = id;
                data.id = id;
                subConfigs.push(config);
                subDatas.push(data);
            })) return;

            // 解析导出变量名
            let vname = '', exportName = `__EASYV_COMPONENT__`;
            const ast = parser(content, { sourceType: 'module', plugins: ['jsx', 'classProperties'] });
            traverse(ast, {
                ExportDefaultDeclaration(path) {
                    const child = path.node.declaration;

                    if (child.type === 'ClassDeclaration') {
                        // e.g: export default class a {}
                        if (child.id) {
                            vname = child.id.name;
                            // e.g: export default class {}
                        } else {
                            vname = exportName;
                            child.id = t.identifier(exportName);
                        }
                        // e.g: export default memo() 
                        // e.g: export default () => {} 
                    } else if (!!~['CallExpression', 'ArrowFunctionExpression', 'FunctionExpression'].indexOf(child.type)) {
                        const exportIdentifier = t.identifier(exportName);
                        path.insertBefore(t.variableDeclaration('const', [t.variableDeclarator(exportIdentifier, child)]));
                        path.node.declaration = exportIdentifier;
                        vname = exportName;
                        // e.g: export default function(){} 
                    } else if (!!~['FunctionDeclaration'].indexOf(child.type)) {
                        const exportIdentifier = t.identifier(exportName);
                        child.id = exportIdentifier;
                        vname = exportName;
                        // e.g: export default a;
                    } else if (child.type === 'Identifier') {
                        vname = child.name
                    }
                    path.stop();
                },
                ExportNamedDeclaration(path) {
                    const child = path.node.declaration;
                    // e.g: export const a = {};
                    if (child.type === 'VariableDeclaration') {
                        const exportIdentifier = child.declarations[0].id;
                        vname = exportIdentifier.name;
                        path.replaceWith(child);
                        path.insertAfter(t.exportDefaultDeclaration(exportIdentifier))
                        // e.g: export function a(){};
                    } else if (child.type === 'FunctionDeclaration') {
                        const exportIdentifier = child.id;
                        path.replaceWith(child);
                        path.insertAfter(t.exportDefaultDeclaration(exportIdentifier))
                    }
                    path.stop();
                }
            })
            file = generate(ast, { quotes: 'single', retainLines: false }).code;
            if (!vname) {
                return callback(`cannot found default export object in ${path.join(modulePath, 'index')}`)
            }

            // 检查width/height
            const mainConfig = main.config;
            // if (!('width' in mainConfig) || !('height' in mainConfig)) {
            //     return callback(`main.json lack field 'width' or 'height'.`)
            // }

            // 添加width/height属性兼容老版本
            let updateChart = false;
            const hasUnderline = mainConfig.configuration[0] ? ('_value' in mainConfig.configuration[0]) : false;
            const keys = {
                name: hasUnderline ? '_name' : 'name',
                displayName: hasUnderline ? '_displayName' : 'displayName',
                value: hasUnderline ? '_value' : 'value',
                type: hasUnderline ? '_type' : 'type',
                children: hasUnderline ? '_children' : 'children',
            }
            mainConfig.configuration = mainConfig.configuration.map(item => {
                if (item[keys.name] === 'chart') {
                    let values = item._value || item.value;
                    values = values.map(item => {
                        if (item[keys.name] === 'dimension') {
                            updateChart = true;
                            return item;
                            // return getOldDimension(mainConfig.width, mainConfig.height, hasUnderline)
                        }
                        return item;
                    })
                    if (!updateChart) {
                        values.unshift(getOldDimension(mainConfig.width, mainConfig.height, hasUnderline))
                        updateChart = true;
                    }
                    item[keys.value] = values;
                    return item;
                }
                return item;
            })
            if (!updateChart) {
                const chart = { [keys.name]: 'chart', [keys.displayName]: '组件', [keys.value]: [getOldDimension(mainConfig.width, mainConfig.height, hasUnderline)] };
                mainConfig.configuration.unshift(chart);
            }

            if (query.isBuild) {
                // 兼容平台子组件配置
                if (subConfigs.length) {
                    const defineMap = subConfigs.reduce((a, b) => {
                        a[b.base.module_name] = b;
                        return a;
                    }, {})
                    const defineChildren = mainConfig.children || subConfigs.map(item => item.base.module_name);
                    const presetChidlren = mainConfig.presetChildren || [];

                    const children = {
                        [keys.name]: 'children',
                        [keys.displayName]: '子组件',
                        [keys.value]: [
                            {
                                [keys.name]: 'component',
                                [keys.displayName]: '管理',
                                [keys.value]: [
                                    {
                                        [keys.name]: 'list',
                                        [keys.displayName]: '',
                                        [keys.value]: presetChidlren.filter(id => defineMap[id]).map(id => {
                                            const { base } = defineMap[id];
                                            return { label: base.name, name: base.module_name }
                                        }),
                                        [keys.type]: 'children',
                                        [keys.children]: defineChildren.filter(id => defineMap[id]).map(id => {
                                            const { base } = defineMap[id];
                                            return { name: base.name, value: base.module_name }
                                        })
                                    }
                                ]
                            }
                        ]
                    }
                    mainConfig.configuration.push(children)
                }
                // 替换modulename
                if (query.uniqueId) {
                    mainConfig.base.module_name += '.' + query.uniqueId;
                }
                // 删除平台无法识别的参数
                delete mainConfig.children;
                delete mainConfig.presetChildren;
            }
            file += '\n';
            file += `${vname}.getDefaultConfig=function(){return ${JSON.stringify(mainConfig)}};`;
            file += `${vname}.getDefaultData=function(){return ${JSON.stringify(main.data)}};`;
            file += `${vname}.getDefaultChildrenConfig=function(){return ${JSON.stringify(subConfigs)}};`;
            file += `${vname}.getDefaultChildrenData=function(){return ${JSON.stringify(subDatas)}};`;
            callback(null, file)

        } catch (e) {
            callback(e)
        }


    })

};